/*
 * AbsoluteDjControler.cpp
 *
 * Created: 4/19/2014 7:05:35 PM
 *  Author: Daniel Solis
 
 
 * MasterBrd3SourceCode by  Daniel Solis is licensed under a Creative Commons Attribution NonCommercial-ShareAlike 4.0 International License.

 http://creativecommons.org/licenses/by-nc-sa/4.0/
 https://i.creativecommons.org/l/by-nc-sa/4.0/88x31.png
 http://creativecommons.org/ns#
 
 *Any person with a copy of this Software is allowed to use, copy, modify, merge, publish and
 *distribute the Software, subject to the following conditions:
 *The above permission notice shall be included in all copies or substantial portions of the Software.

 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.

 

 */ 

#define F_CPU 16000000UL
void _reboot_Teensyduino_(void) __attribute__((noreturn));
void _restart_Teensyduino_(void) __attribute__((noreturn));

#include <util/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sfr_defs.h>
#include <stdlib.h>

#define nop asm volatile("nop\n\t"::)
#define sleep asm volatile("sleep\n\t"::)

#include "lib/mydefs.h"
#include "lib/encoder.h"
#include "lib/ADC.h"
#include "lib/SPI.h"
#include "lib/USART.h"
#include "lib/usb_midi/usb_api.h"
#include "lib/usb_midi/usb_private.h"

/*  SUBROUTINES  */
void updateenc_brow(void);
void updateenc_jog1(void);
void updateenc_jog2(void);

/*  FEEDBACK DATA  */
volatile uint8_t ledScan = 0x08;
volatile uint8_t meterDeckA = 0x55;
volatile uint8_t meterDeckB = 0xAA;
volatile uint8_t lowNibbDeckL = 0;
volatile uint8_t highNibbDeckL = 0;
volatile uint8_t lowNibbDeckR = 0;
volatile uint8_t highNibbDeckR = 0;

/*  MATRIX SEMAPHORES  */
volatile uint8_t matrixReadReady_1 = 0;
uint8_t matrixReadReady_2 = 0;
uint8_t matrixProcessReady = 1;

/*  ENCODER DATA  */
Encoder browse;
Encoder jog1;
Encoder jog2;

int main(void)
{
	/*  ANALOG DATA  */
	const uint8_t analogChannelsStatMux1[16] = {1,1,1,1,1,1,1,1,/* */1,1,1,1,1,1,1,1};
	const uint8_t analogChannelsStatMux2[8] = {1,1,1,1,1,1,1,1};
	const uint8_t mux2Selectors[8] = {0, 1, 4, 5, 8, 9, 12, 13};
	uint8_t analogPrevValuesMux1[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	uint8_t analogPrevValuesMux2[8] = {0,0,0,0,0,0,0,0};
	uint8_t muxControl1 = 16;
	uint8_t muxControl2 = 8;
	uint8_t muxChange = 0;
	uint8_t analogValueScaled;
	
	/*  MATRIX DATA  */
	uint64_t butPrev = 0;
	uint64_t but = 0;
	uint64_t butstates = 0;
	uint64_t prevbutstates = 0;
	uint64_t turnOn;
	uint64_t turnOff;
	uint64_t scan2;
	uint8_t buttonShift = 0;
	uint8_t cnt;
	uint8_t dd;
	uint8_t scan;
	
	
	/*  START MCU  */
	CLKPR = 0x80; //NO PRESCALER
	CLKPR = 0x00;
	cli(); //DISABLE INTERRUPTS
	
	/*  POWER REDUCTION  */
	//SMCR = 0x00; //IDLE SLEEP MPDE
	PRR0 = _BV(PRTWI); //DISABLE TWI
	PRR1 = _BV(PRTIM3)/*| _BV(PRUSART1)*/; //DISABLE TIMER3 USART
	
	/*  START IO  */
// 	DDRB = 0;
// 	DDRC = 0;
// 	DDRD = 0;
// 	DDRE = 0;
// 	DDRF = 0;
// 	PORTB = 0;
// 	PORTC = 0;
// 	PORTD = 0;
// 	PORTE = 0;
// 	PORTF = 0;
	
	/*---------------------
	Other:
	ALT1 PC7
	ALT2 PB6
	ALT3 PB7
	ALT4 PC6
	ALT5 PB0
	SHIFT PE6
	Shift registers:
	STCLKLEDS 12 595 ALT5 PB0
	STCLKSEND 12 595 ALT3 PB7
	STCLKRECE 12 597 ALT4 PC6
	PLRECEIVE 13 597 ALT1 PC7
	---------------------*/
	DDRE |= ( _BV(SHIFT) ); //INPUT
	PORTE |= _BV(SHIFT); //PULLUP

	DDRB |= _BV(0) | _BV(7) | _BV(6);
	PORTB &= ~(_BV(0) | _BV(7) | _BV(6));
	DDRC |= _BV(6) | _BV(7);
	PORTC &= ~(_BV(6) | _BV(7));
	/*---------------------
	
	Analog mux 1:
	A 10 MUX1A PD4
	B 11 MUX1B PD5
	C 14 MUX1C PD6
	D 13 MUX1D PD7
	Analog mux 2:
	A 11 MUX1A PD4
	B 10 MUX1C PD6
	C 09 MUX1D PD7
	---------------------*/
 	DDRD |= ( _BV(MUX1A) | _BV(MUX1B) | _BV(MUX1C) | _BV(MUX1D) ); //OUTPUTS
 	PORTD &= ~( _BV(MUX1A) | _BV(MUX1B) | _BV(MUX1C) | _BV(MUX1D) ); //LOW
	
	/*---------------------
	SPI:
	SERDATLEDS 14 595 MOSI PB2
	SERCLKLEDS 11 595 SCLK PB1
	SERDATSEND 14 595 MOSI PB2
	SERCLKSEND 11 595 SCLK PB1
	SERDATRECE 09 597 MISO PB3
	SERCLKRECE 11 597 SCLK PB1
	---------------------*/
 	DDRB |= ( _BV(SS) | _BV(SCLK) | _BV(MOSI) ); //OUTPUTS
// 	PORTB &= ~( _BV(SS) | _BV(SCLK) | _BV(MOSI) ); //LOW
	/*---------------------*/
	
	/*---------------------
	Encoders:
	ENC1A PD0
	ENC1B PD1
	ENC2A PD2
	ENC2B PD3
	SRCH_ENC_A PB4
	SRCH_ENC_B PB5
	---------------------*/
	DDRD &= ~( _BV(ENC1A) | _BV(ENC1B) | _BV(ENC2A) | _BV(ENC2B) ); //INPUTS
	PORTD |= ( _BV(ENC1A) | _BV(ENC1B) | _BV(ENC2A) | _BV(ENC2B) ); //PULLUPS
	EIMSK = _BV(INT0) | _BV(INT1) | _BV(INT2) | _BV(INT3); //Enable external interrupts
	EICRA = _BV(ISC00) | _BV(ISC10) | _BV(ISC20) | _BV(ISC30);
	DDRB &= ~( _BV(SRCH_ENC_A) | _BV(SRCH_ENC_B) ); //INPUTS
	PORTB |= ( _BV(SRCH_ENC_A) | _BV(SRCH_ENC_B) ); //PULLUPS
	PCICR = _BV(PCIE0); //Enable pin change interrupts
	PCMSK0 = _BV(PCINT4) | _BV(PCINT5);
	/*---------------------
	LEDS:
	TR0 PF4
	TR1 PF5
	TR2 PF6
	TR3 PF7
	---------------------*/
 	DDRF |= ( _BV(TR0) | _BV(TR1) | _BV(TR2) | _BV(TR3) ); //OUTPUTS
 	PORTF &= ~( _BV(TR0) | _BV(TR1) | _BV(TR2) | _BV(TR3) ); //LOW
	
	/*---------------------
	Timer0:
	F = 300Hz
	---------------------*/
	TCCR0A = _BV(WGM01); //CTC Top OCR0A
	TCCR0B = _BV(CS02); //N = 256
	OCR0A = 103; //f = FCPU/(2*N*(1+OCR0A))
	TIMSK0 = _BV(OCIE0A);
	
	/*---------------------
	Timer1:
	T 20ms
	---------------------*/
	//TCCR1B = 0;
	TCCR1B = _BV(WGM12);
	OCR1A = 2499; //t = 2*N*(OCR1A+1)/FCPU
	TIMSK1 = _BV(OCIE1A);
	//Timer stopped, started after first matrix reading
	
	/*  START COM  */
	usb_init();
	SPIstart();
	SPIfast();
	ADCstart();
	ADCinput(0);
	//USART serial;
	//serial.init(300000);
	_delay_ms(100);
	
	PCIFR = _BV(PCIF0); //Clear pin change interrupt flag
	EIFR = _BV(INTF0) | _BV(INTF1) | _BV(INTF2) | _BV(INTF3); //Clear external interrupt flag
	
	sei();
	
    while(1)
    {
		PORTB ^= _BV(6);
 		if (muxChange == 0)
 		{
 			ADCinput(0);
 			muxControl1--;	
 			if (analogChannelsStatMux1[muxControl1]) //channel activated
 			{
 				PORTD &= 0x0F;
 				PORTD |= muxControl1 << 4;
 				nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
 				ADCSRA |= _BV(ADSC); //start ADC conversion
 			}
 		}
 		else
 		{
 			ADCinput(1);
 			muxControl2--;
 			if (analogChannelsStatMux2[muxControl2]) //channel activated
 			{
 				PORTD &= 0x0F;
 				PORTD |= mux2Selectors[muxControl2] << 4;
 				nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
 				ADCSRA |= _BV(ADSC); //start ADC conversion
 			}
 		}
 			/* ---------------------------*/
 		
 		
 		
 		/* --------------------------------------- 
 		Browse encoder
 		--------------------------------------- */
 		if (browse.posit != browse.positprev)
 		{
 			if (browse.posit > browse.positprev)
 			{
 				browse.change4 = browse.posit - browse.positprev;
 				if(browse.change4 >= 4)
 				{
 					usbMIDI.sendControlChange(1,0x3F,3);
 					browse.positprev = 12;
 					browse.posit = 12;
 				}
 			}
 			else if (browse.posit < browse.positprev)
 			{
 				browse.change4 = browse.positprev - browse.posit;
 				if(browse.change4 >= 4)
 				{
 					usbMIDI.sendControlChange(1,0x41,3);
 					browse.positprev = 12;
 					browse.posit = 12;
 				}
 			}
 			
 		}
 		/* --------------------------------------- */
 		
 		/* --------------------------------------- 
 		Jog1 encoder
 		--------------------------------------- */
 		if (jog1.posit != jog1.positprev)
 		{
 			if (jog1.posit > jog1.positprev)
 			{
 				usbMIDI.sendControlChange(2,0x3F,3);
 			}
 			else if (jog1.posit < jog1.positprev)
 			{
 				usbMIDI.sendControlChange(2,0x41,3);
 			}
 			jog1.positprev = jog1.posit;
 		}
 		/* --------------------------------------- */
 		
 		/* --------------------------------------- 
 		Jog2 encoder
 		--------------------------------------- */
 		if (jog2.posit != jog2.positprev)
 		{
 			if (jog2.posit > jog2.positprev)
 			{
 				usbMIDI.sendControlChange(3,0x3F,3);
 			}
 			else if (jog2.posit < jog2.positprev)
 			{
 				usbMIDI.sendControlChange(3,0x41,3);
 			}
 			jog2.positprev = jog2.posit;
 		}
 		/* --------------------------------------- */
 		
 		/* --------------------------------------- 
 		Matrix reading 1 
 		--------------------------------------- */
  		if (matrixProcessReady)
  		{
  			scan = 1;
  			dd = 0;
  			do
  			{
 				cli();
 				SPIfast(); //spi fast mode
 				PORTB &= ~(_BV(7)); // latch low
 				SPItransfer(scan); //send data to sp shift register
 				PORTB |= _BV(7); // latch high
 				SPIslow(); //spi slow mode
 				butPrev <<= 8;
 				PORTC |= _BV(6); // latch high
 				PORTC &= ~_BV(6); // latch low
 				PORTC &= ~_BV(7); // pl low
 				PORTC |= _BV(7); //pl high
 				butPrev |= SPItransfer(0); //store byte from P/S shift register
 				sei();
 				scan <<= 1; //set next bit of S/P shift register high
 				dd++;
 				//_delay_ms(1000);
  			} while (dd != 8); //repeat 8 times
  			matrixProcessReady = 0;
  			
  			TCNT1 = 0; //Restart timer
  			TIFR1 |= _BV(OCF1A); //Clear flag
  			TIMSK1 |= _BV(OCIE1A); //Enable interrupt
  			TCCR1B |= ( _BV(CS11) | _BV(CS10) ); //N = 64
 			
  		}
  		/* --------------------------------------- */
  		
  		
  		/* --------------------------------------- 
  		Matrix reading 2
  		--------------------------------------- */
  		if (matrixReadReady_1)
  		{	
  			scan = 1;
  			dd = 0;
  			do
 			{
 				cli();
 				SPIfast();
 				PORTB &= ~(_BV(7)); // latch low
 				SPItransfer(scan); //send data to sp shift register
 				PORTB |= _BV(7); // latch high
 				SPIslow();
 				but <<= 8;
 				PORTC |= _BV(6); // latch high
 				PORTC &= ~_BV(6); // latch low
 				PORTC &= ~_BV(7); // pl low
 				PORTC |= _BV(7); //pl high
 				but |= SPItransfer(0);
 				sei();
 				scan <<= 1;
 				dd++;
				 //_delay_ms(1000);
  			} while (dd != 8);
  			matrixReadReady_1 = 0;
  			matrixReadReady_2 = 1;
   		}
  		/* --------------------------------------- */
  		
  		
  		/* --------------------------------------- 
  		Matrix process
  		--------------------------------------- */
  		if (matrixReadReady_2)
  		{
 			if (but == butPrev) //both readings are the same
 			{ 
 				butstates = but; //valid data
 			}
 			
 			if (butstates != prevbutstates) //any button changed state
 			{
 				turnOn = ~prevbutstates & butstates; //buttons to be turned on
 				turnOff = prevbutstates & ~butstates; //buttons to be turned off
 				
 				
 				if(turnOn) //there is something to turn on
 				{
 					scan2 = 1;
 					cnt = 0;
 					do
 					{
 						if (turnOn & scan2) //button needs to be turned on
 						usbMIDI.sendNoteOn((cnt + buttonShift), 127, 1);
 						scan2 <<= 1; //compare next bit
 						cnt++;
 					} while (cnt != 64); //repeat 64 times
 				}
   				
   				if(turnOff) //there is something to turn off
   				{
   					scan2 = 1;
   					cnt = 0;
   					do
   					{
   						if (turnOff & scan2) //button needs to be turned off
   						usbMIDI.sendNoteOff((cnt + buttonShift), 127, 1);
   						scan2 <<= 1;
   						cnt++;
   					} while (cnt != 64); //repeat 64 times
   				}
 			}
  			prevbutstates = butstates; //store current state of buttons
  			matrixReadReady_2 = 0;
  			matrixProcessReady = 1;
  		}
  		/* --------------------------------------- */
  		
  		
  		
  		/* --------------------------------------- 
  		Data in
  		--------------------------------------- */		
  		if(usbMIDI.read())
  		{
  			uint8_t type = usbMIDI.getType();
  			uint8_t note = usbMIDI.getData1();
  			uint8_t velo = usbMIDI.getData2();
  			uint8_t chan = usbMIDI.getChannel();
  			
  			uint16_t temp = 0;
  			uint8_t eqval = 0;
  			switch(chan) 
  			{
  				case 1:
  					if(type == 3)
  					{
  						if(note == 0 || note == 1)
  						{
 //  							if(velo > 49)
 //  							{
 //  								temp = velo / 10;
 //  								velo = temp - 5;
 //  								temp = 1 << velo;
 //  								velo = temp;
 //  								if(temp > 1)
 //  								temp |= (velo - 1);
 //  							}
 //  							else
 //  								temp = 0x01;
 							temp = (velo * 8)/127;
 							eqval = 0xFF >> (8-(uint8_t)temp);
  							if (note == 0)
  								meterDeckA = eqval;
  							else if (note == 1)
  								meterDeckB = eqval;
  						}
  					}
  					break;
  				
  				case 2:
  					switch (type)
  					{
  						case 0: //note off
  						if (note < 8)
  							lowNibbDeckL &= ~(1 << note);
  						else
  						{
  							note -= 8;
  							highNibbDeckL &= ~(1 << note);
  						}
  						break;
  						
  						case 1: //note on
  						if (note < 8)
  							lowNibbDeckL |= (1 << note);
  						else
  						{
  							note -= 8;
  							highNibbDeckL |= (1 << note);
  						}
  						break;
  					}
  					break;
  					
  				case 3:
  					switch (type)
  					{
  						case 0: //note off
  						if (note < 8)
  						lowNibbDeckR &= ~(1 << note);
  						else
  						{
  							note -= 8;
  							highNibbDeckR &= ~(1 << note);
  						}
  						break;
  					
  						case 1: //note on
  						if (note < 8)
  						lowNibbDeckR |= (1 << note);
  						else
  						{
  							note -= 8;
  							highNibbDeckR |= (1 << note);
  						}
  						break;
  					}
  					break;
  			}
  		}
 		/* --------------------------------------- */
 		if (muxChange == 0)
 		{
 			while(ADCSRA & (1<<ADSC));
 			if (analogChannelsStatMux1[muxControl1]) //channel activated
 			{
 				analogValueScaled = ADCH;	//get adc value
 				analogValueScaled >>= 1;
 				if ( ( abs(analogPrevValuesMux1[muxControl1] - analogValueScaled) ) > 3 ) //control value change is greater than threshold
 				{
 					//analogValueScaled = (analogValue*127)/255;
 					usbMIDI.sendControlChange(muxControl1, analogValueScaled, 1); //send midi data
 					analogPrevValuesMux1[muxControl1] = analogValueScaled; //save analog value
 				}
 			}
 			if (muxControl1 == 0)
 			{
 				muxControl1 = 16;
 				muxChange = 1;
 			}
 		}
 		else
 		{
 			while(ADCSRA & (1<<ADSC));
 			if (analogChannelsStatMux1[muxControl2]) //channel activated
 			{
 				analogValueScaled = ADCH;	//get adc value
 				analogValueScaled >>= 1;
 				if ( ( abs(analogPrevValuesMux2[muxControl2] - analogValueScaled) ) > 3 ) //control value change is greater than threshold
 				{
 					//analogValueScaled = (analogValue*127)/255;
 					usbMIDI.sendControlChange((muxControl2+16), analogValueScaled, 1); //send midi data
 					analogPrevValuesMux2[muxControl2] = analogValueScaled; //save analog value
 				}
 			}
 			if (muxControl2 == 0)
 			{
 				muxControl2 = 8;
 				muxChange = 0;
 			}
 		}
		
		//_delay_ms(40);
		//_delay_us(490);
	}
}

ISR(ADC_vect)
{
	
}


ISR(TIMER0_COMPA_vect)
{
	/* --------------------------------------- 
		Led multiplexer  
	--------------------------------------- */
	if ( (!SPItransmitting) && (!(SPSR & _BV(SPIF))) )
 	{
		SPIfast();
		ledScan <<= 1;
		switch(ledScan)
		{
			case 0x10:
 				SPItransfer(meterDeckA);
				break;
			case 0x20:
				SPItransfer(meterDeckB);
				break;
			case 0x40:
				SPItransfer(lowNibbDeckL);
				break;
			case 0x80:
				SPItransfer(lowNibbDeckR);
				break;
			case 0x00:
				ledScan = 0x08;
				break;
		}
		PORTB &= ~_BV(0);
		PORTB |= _BV(0);
		PORTF = (PORTF & 0x0F) | (ledScan);
	}
}

ISR(TIMER1_COMPA_vect)
{
	TCCR1B = _BV(WGM12);
	//PORTB ^= 0x01;
	matrixReadReady_1 = 1;
}

ISR(PCINT0_vect)
{
	updateenc_brow();
}

ISR(INT0_vect)
{
	updateenc_jog1();
}

ISR(INT1_vect)
{
	updateenc_jog1();
}

ISR(INT2_vect)
{
	updateenc_jog2();
}

ISR(INT3_vect)
{
	updateenc_jog2();
}

void updateenc_brow(void)
{
	browse.s = browse.state & 3;
	if (PINB & 0x10) browse.s |= 4;
	if (PINB & 0x20) browse.s |= 8;
	switch ( (browse.s & 0x0F) )
	{
		case 0: case 5: case 10: case 15:
		break;
		case 1: case 7: case 8: case 14:
		browse.posit++;
		break;
		case 2: case 4: case 11: case 13:
		browse.posit--;
		break;
		case 3: case 12:
		browse.posit += 2;
		break;
		default:
		browse.posit -= 2;
		break;
	}
	browse.state = (browse.s >> 2);
}

void updateenc_jog1(void)
{
	jog1.s = jog1.state & 3;
	if (PIND & 0x01) jog1.s |= 4;
	if (PIND & 0x02) jog1.s |= 8;
	switch ( (jog1.s & 0x0F) )
	{
		case 0: case 5: case 10: case 15:
		break;
		case 1: case 7: case 8: case 14:
		jog1.posit++;
		break;
		case 2: case 4: case 11: case 13:
		jog1.posit--;
		break;
		case 3: case 12:
		jog1.posit += 2;
		break;
		default:
		jog1.posit -= 2;
		break;
	}
	jog1.state = (jog1.s >> 2);
}

void updateenc_jog2(void)
{
	jog2.s = jog2.state & 3;
	if (PIND & 0x04) jog2.s |= 4;
	if (PIND & 0x08) jog2.s |= 8;
	switch ( (jog2.s & 0x0F) )
	{
		case 0: case 5: case 10: case 15:
		break;
		case 1: case 7: case 8: case 14:
		jog2.posit++;
		break;
		case 2: case 4: case 11: case 13:
		jog2.posit--;
		break;
		case 3: case 12:
		jog2.posit += 2;
		break;
		default:
		jog2.posit -= 2;
		break;
	}
	jog2.state = (jog2.s >> 2);
}